"""
NthEvent server and utilities

This module provides an event server and client utilities for
MIDI and iGesture devices.
 
Copyright (c) 2007, Tim Thompson
All rights reserved.	

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

-  Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

-  Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.

-  Neither the name of Tim Thompson, nosuch.com, nor the names of
   any contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import sys
import time
import traceback
import thread
import threading
import copy
import asyncore
import asynchat
import socket
import sys
import re
import xml.dom.minidom as xmldom
import string
import pypm
import os.path

from os.path import isdir,isfile,isabs,abspath
from urllib import quote,unquote
from threading import *
from ctypes import *
from time import sleep
from Queue import Queue, Empty
from xml.sax import saxutils
from xml.dom import Node
from traceback import format_exc 
from dircache import listdir

class BaseEvent:
	def __init__(self):
		self.time = 0.0

class FretDownEvent(BaseEvent):

	def __init__(self,strnum,ev):
		BaseEvent.__init__(self)
		self.strnum = strnum
		self.prox = ev.prox
		self.x = ev.x
		self.y = ev.y

	def to_osc(self):
		return("/multitouch/fretdown",[self.strnum])

class FretUpEvent(BaseEvent):

	def __init__(self,strnum):
		BaseEvent.__init__(self)
		self.strnum = strnum

	def to_osc(self):
		return("/multitouch/fretup",[self.strnum])

debug = False

from nosuch.fingerutil import *
from nosuch.midiutil import *
from nosuch.midifile import *
from osc.oscAPI import *

class NthFretMonitor(Thread):

	def __init__(self,es):

		Thread.__init__(self)
		self.eventserver = es
		self.setDaemon(True)

	def run(self):
		while True:
			self.eventserver.fret_check()
			sleep(0.01)


class NthEventServer(Thread):
	"""
	Provides an event stream that can serve multiple listeners
	track of what fingers are currently down, smoothing drag motion, etc.
	"""

	oneServer = None

	def __init__(self):

		Thread.__init__(self)
		self.setDaemon(True)

		self.cv = threading.Condition()
		self.events = {}
		self.firstevent = 0
		self.nextevent = 0
		self.osc_socket = None
		self.osc_addr = None
		self.osc_port = None

		self.start()
		self.too_old_seconds = 30.0
		self.event_inputs = {}
		self.forward_inputs = {}
		self.forward_finger = None
		self.last_finger_event = {}
		self.do_frets = False
		self.fret_isdown = {1:0,2:0,3:0,4:0,5:0,6:0}
		self.fret_lastchange = {1:0,2:0,3:0,4:0,5:0,6:0}
		t = NthFretMonitor(self)
		t.start()

	@staticmethod
	def server():
		if NthEventServer.oneServer == None:
			NthEventServer.oneServer = NthEventServer()
		return NthEventServer.oneServer

	@staticmethod
	def push_event(ev,data):
		NthEventServer.server()._push_event(ev)

	@staticmethod
	def current_index():
		return NthEventServer.server()._current_index()

	@staticmethod
	def num_events_after(lasti):
		return NthEventServer.server()._num_events_after(lasti)

	def set_midi_timezero(self,t):
		self.midi_timezero = t;

	def set_finger_timezero(self,t):
		self.finger_timezero = t;

	def set_midi_event_input(self,i):
		self.event_inputs[i] = 1

	def set_midi_forward(self,i,outs):
		self.forward_inputs[i] = outs

	def set_osc_recipient(self,addr,port):
		self.osc_addr = addr
		self.osc_port = int(port)
		self.osc_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
		print "Sending OSC messages to IP address",addr,"port",port

	def set_finger_forward(self,o):
		self.forward_finger = o

	def set_finger_order(self,o):
		self.finger_order = o

	def set_fingertype(self,t):
		if t == "fret":
			self.do_frets = True
		elif t == "raw":
			self.do_frets = False
		else:
			print "Unrecognized value in fingertype: ",t

	def send_osc(self,msg_addr,msg_data):
		self.osc_socket.sendto(createBinaryMsg(msg_addr,msg_data), (self.osc_addr, self.osc_port))

	def fret_check(self):
		# print "FRET_CHECK ",time.time()
		now = time.time()
		for strnum in range(1,7):
			if not self.fret_isdown[strnum]:
				continue
			dt = now - self.fret_lastchange[strnum]
			if dt > 0.1:
				print "String ",strnum," is going back up (fret_check) ",time.time()
				self.fret_isdown[strnum] = 0

				fev = FretUpEvent(strnum)
				self._send_osc_event(fev)

				self.fret_lastchange[strnum] = now

	def _fret_event(self,ev):
		f = ev.finger
		# print "_fret_event ev=",ev,"  f=",f
		dy = 1.0 / 6.0
		if ev.y < dy:
			strnum = 1
		elif ev.y < 2.0*dy:
			strnum = 2
		elif ev.y < 3.0*dy:
			strnum = 3
		elif ev.y < 4.0*dy:
			strnum = 4
		elif ev.y < 5.0*dy:
			strnum = 5
		else:
			strnum = 6

		if isinstance(ev,FingerUp):
			print "             FINGER EVENT UP!  finger=",ev.finger,"prox",ev.prox,"time",time.time()
		if isinstance(ev,FingerDown):
			print "             FINGER EVENT DOWN!  finger=",ev.finger,"prox",ev.prox,"time",time.time()
		if isinstance(ev,FingerDrag):
			print "             FINGER EVENT DRAG!  finger=",ev.finger,"prox",ev.prox,"time",time.time()

		if isinstance(ev,FingerDrag) or isinstance(ev,FingerDown):
			if not self.fret_isdown[strnum]:
				self.fret_isdown[strnum] = 1
				fev = FretDownEvent(strnum,ev)
				self._send_osc_event(fev)
				print "String",strnum,"JUST GOT PRESSED ",time.time()
			else:
				# if the proximity gets low enough,
				# we consider it a finger-up event
				if ev.prox < 0.25:
					print "PROX LOW, string",strnum,"up!"
					self.fret_isdown[strnum] = 0
					fev = FretUpEvent(strnum)
					self._send_osc_event(fev)
		else:   # FingerUp
			pass
			# We ignore normal fingerup events because they are
			# sometimes spurious (fingerup and fingerdown will
			# come in quick succession).  Instead we rely on
			# fret_check to detect when a string has been let up.

			# if self.fret_isdown[strnum]:
			# 	print "String",strnum,"is going back up (fingerup)",time.time()
			# 	self.fret_isdown[strnum] = 0

		self.fret_lastchange[strnum] = time.time()
		# self._send_osc_event(ev)

	def _push_event(self,ev):

		if isinstance(ev,FingerEvent):
			ev.time -= self.finger_timezero
			eventit = True

			i = ev.devindex
			if i <= 3:
				# The finger_order hack allows the caller
				# of the server to adjust the devindex values
				# returned for finger events
				ev.devindex = ord(self.finger_order[i])-ord("0")

			# Don't send finger drag events (for the same finger)
			# closer than .01 secs apart

			if isinstance(ev,FingerDrag):
				if ev.uid in self.last_finger_event:
					dt = ev.time - self.last_finger_event[ev.uid].time
					if dt < 0.01:
						eventit = False

			if eventit and self.forward_finger:
				f = self.forward_finger
				evtext = "%s" % ev
				m = SysEx(0xf0)
				m.append(0)
				m.append(0x7f)
				for c in evtext:
					m.append(ord(c))
				m.append(0xf7)
				f.schedule(m)

			if eventit:
				if self.osc_socket != None:
					if self.do_frets:
						self._fret_event(ev)
					else:
						self._send_osc_event(ev)
				self.last_finger_event[ev.uid] = ev

		elif isinstance(ev,MidiEvent):
			ev.time -= self.midi_timezero
			eventit = False
			# See if it's one we should forward or event
			m = ev.midimsg
			# Should this be more efficient than a loop?
			# There aren't very many...
			for i in self.forward_inputs:
				if i == m.device:
					for o in self.forward_inputs[i]:
						o.schedule(m)
			for i in self.event_inputs:
				if i == m.device:
					eventit = True
					break
			if self.osc_socket != None:
				self._send_osc_event(ev)
		else:
			print "Unknown event type in _push_event!? ev=",ev
			eventit = False

		if eventit:
			self.cv.acquire()
			# The events we're getting (e.g. from pyportmidi)
			# have a .time value that is more accurate/correct
			# than we could do here.
			self.events[self.nextevent] = ev
			self.nextevent += 1
			global debug
			# if debug:
			# 	print "Added event, len is now ",len(self.events)

			self.cv.notifyAll()

			too_old = ev.time - self.too_old_seconds
			while self.events[self.firstevent].time < too_old \
				and self.firstevent < self.nextevent:

				del self.events[self.firstevent]
				self.firstevent += 1
			self.cv.release()

	def _send_osc_event(self,ev):
		(msg_addr,msg_data) = ev.to_osc()
		self.send_osc(msg_addr,msg_data)

	def _current_index(self):
		return self.nextevent - 1

	def _num_events_after(self,lasti):
		nexti = lasti + 1
		if nexti < self.firstevent:
			return 0
		if nexti > self.nextevent:
			return 0	# shouldn't happen
		return self.nextevent - nexti

	def _event_after(self,lasti,wait=False):
		self.cv.acquire()
		thisi = lasti + 1
		if thisi < self.firstevent:
			thisi = self.firstevent
		if thisi >= self.nextevent:
			if not wait:
				self.cv.release()
				return None

			self.cv.wait()
			if thisi >= self.nextevent:
				print "Hey, wait returned but nextevent hasn't changed!?"
				self.cv.release()
				return None
		e = copy.copy(self.events[thisi])
		self.cv.release()
		return (e,thisi)

class NthEventListener:
	"""
	Provide independent access to a stream of events.
	"""
	def __init__(self):
		# start from the current index
		server = NthEventServer.server()
		self.lastindex = server.current_index()

	def next_event(self,timeout=5.0):
		server = NthEventServer.server()
		arr = server._event_after(self.lastindex,wait=False)
		if arr == None:
			# Wait for the next event
			arr = server._event_after(self.lastindex,wait=True)
			if arr == None:
				print "Hey, waited for event, but got None!?"
				return None
		self.lastindex = arr[1]
		return arr[0]

	def num_events_waiting(self):
		server = NthEventServer.server()
		return server.num_events_after(self.lastindex)

class NthEventMonitor(Thread):
	"""
	A thread that prints all events it sees.
	"""

	monitornum = 0

	def __init__(self):
		Thread.__init__(self)
		self.setDaemon(True)
		self.start()

	def run(self):
		f = NthEventListener()
		while True:
			print f.next_event()

# class NthEventOscMonitor(Thread):
# 	"""
# 	A thread that prints all events it sees.
# 	"""
# 
# 	monitornum = 0
# 
# 	def __init__(self):
# 		Thread.__init__(self)
# 		self.setDaemon(True)
# 		self.start()
# 
# 	def run(self):
# 		print "OSC MONITOR started running..."
# 		f = NthEventListener()
# 		while True:
# 			print "OSC MONITOR loop..."
# 			print f.next_event()

class NthServer (asyncore.dispatcher):

	"""listens for connections"""
	def __init__ (self, host, port,midiin=None,midiout=None,rootdir=None,oscaddr=None,oscport=None,fingertype="raw"):
		asyncore.dispatcher.__init__ (self)
		self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
		self.set_reuse_addr()
		self.bind((host, port))
		self.listen (5)
		self.count = 0
		self.connections = {}
		self.server = NthEventServer.server()

		Midi.startup()
		Finger.startup()

		self.server.set_fingertype(fingertype)
		self.server.set_midi_timezero(Midi.time_now())
		self.server.set_finger_timezero(Finger.time_now())
		if oscaddr != None and oscport != None:
			self.server.set_osc_recipient(oscaddr,oscport)

		self.inputs = {}
		self.outputs = {}
		self.rootdir = rootdir

		Midi.callback(NthEventServer.push_event,"")
		Finger.callback(NthEventServer.push_event,"")

		if midiout:
			self.open_outputs(midiout)
		if midiin:
			self.open_inputs(midiin)

	def open_outputs(self,midiout):
		for nm in MidiOutput.devices():
			if re.search(midiout,nm):
				print("Opening Output = "+nm)
				o = MidiOutput(nm)
				try:
					o.open()
					self.outputs[nm] = o
				except:
					self.logerror("Error during open of output %s - %s"%(nm,format_exc()))
					continue

	def open_inputs(self,midiin):
		for nm in MidiInput.devices():
			if re.search(midiin,nm):
				print("Opening Input = "+nm)
				i = MidiInput(nm)
				try:
					i.open()
					self.inputs[nm] = i
				except:
					self.logerror("Error during open of input %s - %s"%(nm,format_exc()))
					continue

	def midi_forward(self,in_expr,out_expr):

		# Find the real outputs matched by out_expr
		outs = []
		for nm in self.outputs:
			if re.search(out_expr,nm):
				outs.append(self.outputs[nm])

		if len(outs) <= 0:
			print "Hey, no outputs matched: ",out_expr
			return

		any = False
		for nm in self.inputs:
			# print "looking at nm=",nm," in_expr=",in_expr
			if re.search(in_expr,nm):
				any = True
				self.server.set_midi_forward(self.inputs[nm],outs)
		if not any:
			print "Hey, no inputs matched: ",in_expr

	def finger_forward(self,out_expr):
		o = None
		# Find the real name matched by the regular expression
		for nm in self.outputs:
			if re.search(out_expr,nm):
				o = self.outputs[nm]
				break
		if o:
			self.server.set_finger_forward(o)

	def finger_order(self,o):
		self.server.set_finger_order(o)

	def midi_events(self,in_expr):
		i = None
		# Find the real names matched by the regular expression
		for nm in self.inputs:
			if re.search(in_expr,nm):
				self.server.set_midi_event_input(self.inputs[nm])

	def handle_accept (self):
		self.count += 1
		try:
			NthResponder (self, self.count, self.rootdir, self.accept())
		except:
			self.logerror("accept() error on connection %d - %s"%(self.count,format_exc()))

	def logerror(self,msg):
		print("NthServer error: "+msg);

	def destroy (self):
		Midi.shutdown()
		asyncore.socket_map.clear()
		raise asyncore.ExitNow

	def shutdown(self):
		self.destroy()

class NthServerThread(Thread):
	def __init__(self):
		Thread.__init__(self)

	def run(self):
		try:
			asyncore.loop(1.0)
		except asyncore.ExitNow:
			pass

class NthResponder (asynchat.async_chat):
	"""this is what does the server<=>client stuff"""
	def __init__ (self, server, count, rootdir, (conn, addr)):
		asynchat.async_chat.__init__(self, conn)
		self.readqueue = Queue(0)
		self.set_terminator ('\0')
		self.beingClosed = False
		self.server = server
		self.buffer = ''
		self.id = id(self)
		self.ip = addr[0]
		self.count = count
		self.rootdir = rootdir
		# initialized on connect
		self.service = None
		self.server.connections[self.id]=self
		self.execthread = thread.start_new_thread(self.module_executor,(self,))
		self.currentnode = None
		self.shutdownOnClose = False
		self.listener = NthEventListener()
		self.lock = thread.allocate_lock()

	def collect_incoming_data (self, data):
		self.buffer += data
    
	def found_terminator (self):
		"""we just received a nice document, do stuff"""
		self.readqueue.put(self.buffer)
		self.buffer = ''

	def safeshutdown(self):
		self.shutdownOnClose = True
		self.send_ok()

	def quickshutdown(self):
		print("Starting quick shutdown")
		self.send_ok()
		self.server.shutdown()

	def logerror(self,msg):
		self.server.logerror(msg)

	def module_executor(self,responder):
	    try:
		while True:
			data = self.readqueue.get()
			self.execute_doc(data)
	    except:
		e = format_exc()
		self.logerror("Error in module_executor error=%s" % (e))

	def execute_doc(self,data):
		global debug
		if debug:
			print("data=",data)
		try:
			doc = xmldom.parseString(data)
		except:
			self.logerror("Exception: "+format_exc())
			self.logerror("XML data that caused it: "+str(data))
			return

		for x in doc.childNodes:
			self.execute_xml(x)
		doc.unlink()

	def execute_xml(self,x):
		global debug
		try:
			self.onXML(x)
		except:
			e = format_exc()
			self.logerror("Error in module for node=%s error=%s"
					% (x.localName,e))

	def send_error(self,e):
		"""sends an <error value="%s" /> tag to the client
		"""
		xml = ['<error>%s</error>' % saxutils.escape(e)]
		self.send_response(string.join(xml,''))

    
	def handle_close (self):
		print("Handle_close!")
		self.beingClosed = True
		try:
			if self.server.connections.has_key(self.id):
				del(self.server.connections[self.id])
			self.close()
		except:
			e = format_exc()
			print("UNEXPECTED error in handle_close: %s" % (e))

		if self.shutdownOnClose:
			self.server.shutdown()

	def handle_write(self):
		# The asyncore.loop thread calls this.
		self.lock.acquire()
		asynchat.async_chat.handle_write(self)
		self.lock.release()

	def destroy (self):
		if self.beingClosed:
			return
		self.handle_close()

	def send_response(self,data):
		if self.beingClosed:
			return

		# Because there are two threads that are accessing
		# the async_chat producer fifo (this one, and the
		# main asyncore loop thread), we need to lock things.
		global debug
		if debug:
			print("send_response data=",data)

		self.lock.acquire()
		try:
			asynchat.async_chat.push(self,data+self.terminator)
		except:
			self.logerror("send_response error: "+format_exc())
		self.lock.release()

	def onXML (self,doc):
		global debug
		if debug:
			print("onXML=",doc.toxml())
		try:
			if doc.localName == "next_event": self.next_event(doc)
			elif  doc.localName == "logout": self.handle_close()
			elif  doc.localName == "quickshutdown": self.quickshutdown()
			elif  doc.localName == "shutdown": self.safeshutdown()
			elif  doc.localName == "list_devices": self.list_devices()
			elif  doc.localName == "list_dir": self.list_dir(doc)
			elif  doc.localName == "read_file": self.read_file(doc)
			elif  doc.localName == "write_file": self.write_file(doc)
			elif  doc.localName == "midi_playfile": self.midi_playfile(doc)
			elif  doc.localName == "list_midi_inputs": self.list_midi_inputs()
			elif  doc.localName == "list_midi_outputs": self.list_midi_outputs()
			elif  doc.localName == "write_midi": self.write_midi(doc)
			else:  self.send_error("No such method: "+doc.localName)
		except:
			e = "Error while handling '%s' : %s" % (doc.localName,format_exc())
			self.send_error(e)
			self.logerror(e)

	def check_path(self, path):
		if self.rootdir == None:
			self.send_error("NthServer does not have rootdir configured.");
			return None
		fullpath = os.path.join(self.rootdir,path)
		normpath = abspath(fullpath)
		if not normpath.startswith(self.rootdir):
			self.send_error("NthServer does not allow access outside the rootdir.")
			return None
		return fullpath

	def send_ok(self):
		xml = ["<ok/>"]
		self.send_response(string.join(xml,''))

	def midi_playfile (self, doc):
		dirname = doc.getElementsByTagName("name")[0].childNodes[0].data
		path = self.check_path(dirname)
		if path == None:
			return
		print "midi_playfile, path=",path
		p = Phrase.fromMidiFile(path)
		print "path=",path,"  Length of Phrase is ",len(p.events)
		outputs = self.server.outputs
		for m in p.events:
			for nm in outputs:
				# print "Sending m=",m," to output=",nm
				o = outputs[nm]
				o.schedule(m)

		self.send_ok()

	def list_dir(self,doc):
		if self.rootdir == None:
			self.send_error("NthServer does not have rootdir configured.");
			return

		dirname = doc.getElementsByTagName("name")[0].childNodes[0].data
		path = self.check_path(dirname)
		if path == None:
			return
		if not isdir(path):
			self.send_error("Not a directory: %s" % (dirname))
			return
		list = listdir(path)
		xml = ["<directory>"]
		for f in list:
			listpath = os.path.join(path,f)
			escapef = saxutils.escape(f)
			if isdir(listpath):
				xml.append("<dir>"+escapef+"</dir>")
			elif isfile(listpath):
				xml.append("<file>"+escapef+"</file>")
		xml.append("</directory>")
		self.send_response(string.join(xml,''))

	def read_file(self,doc):
		if self.rootdir == None:
			self.send_error("NthServer does not have rootdir configured.");
			return
		fname = doc.getElementsByTagName("name")[0].childNodes[0].data
		path = self.check_path(fname)
		if path == None:
			return
		try:
			f = open(path,mode="rb");
			d = f.read();
			f.close();
			xml = ["<file_contents>"]
			# Convert all non-printable chars to %xx form
			d = quote(d);
			xml.append(saxutils.escape(d));
			xml.append("</file_contents>")
			self.send_response(string.join(xml,''))
		except:
			e = "Error while reading '%s' : %s" % (fname,format_exc())
			self.send_error(e)
			self.logerror(e)

	def write_file(self,doc):
		if self.rootdir == None:
			self.send_error("NthServer does not have rootdir configured.");
			return
		fname = doc.getElementsByTagName("name")[0].childNodes[0].data
		data = doc.getElementsByTagName("data")[0].childNodes[0].data
		path = self.check_path(fname)
		if path == None:
			return

		try:
			f = open(path,mode="wb");
			data = unquote(data);
			f.write(data);
			f.close();
			self.send_ok()
		except IOError, e:
			# Don't give away full path info in the error
			self.send_error("Unable to write file.");
			self.logerror(format_exc())
		except:
			self.send_error("Unexpected error while writing file.");
			self.logerror(format_exc())

	def list_midi_inputs(self):
		xml = ["<midi_inputs>"]
		for nm in self.server.inputs:
			xml.append(self.server.inputs[nm].to_xml());
		xml.append("</midi_inputs>")
		self.send_response(string.join(xml,''))

	def list_midi_outputs(self):
		xml = ["<midi_outputs>"]
		for nm in self.server.outputs:
			xml.append(self.server.outputs[nm].to_xml());
		xml.append("</midi_outputs>")
		self.send_response(string.join(xml,''))

	def write_midi(self,doc):
		for node in doc.childNodes:
			if node.nodeType == Node.ELEMENT_NODE:
				m = MidiMsg.from_xml(node)
				attrs = node.attributes
				a = attrs.get("devindex")
				if a == None:
					# if no explicit devindex,
					# send it to all outputs
					for nm in self.server.outputs:
						o = self.server.outputs[nm]
						o.schedule(m)
				else:
					devindex = int(a.nodeValue)
					o = self.find_output_by_index(devindex)
					o.schedule(m)
		self.send_ok()

	def find_output_by_index(self,i):
		for nm in self.server.outputs:
			if self.server.outputs[nm].index == i:
				return self.server.outputs[nm]
		return None

	def list_devices(self):

		xml = ["<devices>"]

		devs = MidiInput.devices() 
		for nm in devs:
			o = MidiInput(nm)
			xml.append(o.to_xml())

		devs = MidiOutput.devices() 
		for nm in devs:
			o = MidiOutput(nm)
			xml.append(o.to_xml())

		for n in range(Finger.num_devices()):
			o = FingerDevice(n)
			xml.append(o.to_xml())

		xml.append("</devices>")
		self.send_response(string.join(xml,''))

	def next_event(self,doc):
		"""Returns an xml document containing the next event"""
		# This function typically blocks...
		# print("TJT DEBUG, next_event, doc=",doc.toxml())
		attrs = doc.attributes
		a = attrs.get("overloaded")
		if a and a.nodeValue == "true":
			while True:
				e = self.listener.next_event()
				if isinstance(e,FingerDrag):
					# When we're overloaded, ignore all
					# FingerDrag events
					continue
				if isinstance(e,MidiEvent) and isinstance(e.midimsg,ChanMsg):
					if e.midimsg.channel != 16:
						# We're overloaded, so we
						# ignore everything except
						# the MIDI control channel.
						# There should be an API
						# for setting this rather than
						# harcoding 16
						continue
				break
			# print("OVERLOADED IS FINALLY RETURNING e=",e.to_xml())
		else:
			e = self.listener.next_event()

		self.send_response(e.to_xml())
        
class NthClient (asynchat.async_chat):

	def __init__ (self, host, port):
		asynchat.async_chat.__init__ (self)
		self.create_socket (socket.AF_INET, socket.SOCK_STREAM)
		self.terminator = '\0'
		self.connected = False
		self.host = host
		self.port = port
		self.connect((host,port))
		self.buffer = ''

	def collect_incoming_data (self, data):
		self.buffer += data
    
	def found_terminator (self):
		data=self.buffer
		self.buffer = ''
		self.onData(data)

	def handle_error (self):
		if not self.connected:
			self.onError("Unable to connect to %s:%d" % (self.host,self.port))
		else:
			self.onError("Unknown error in handle_error")
		self.handle_close()

	def handle_expt (self):
		self.handle_error()

	def handle_close (self):
		self.onClose()
		self.close()

	def handle_connect (self):
		self.connected = True
		self.onConnect()

	def push(self, data):
		"pushes data to the server, null termination is automatic"
		asynchat.async_chat.push(self, data+self.terminator)

class NthClientThread(Thread):

	def __init__(self):
		Thread.__init__(self)
		self.setDaemon(True)

	def run(self):
		try:
			asyncore.loop()
		except asyncore.ExitNow:
			# normal exit
			print("NthClientThread got asyncore.ExitNow!!")
			pass
		except:
			print("Unexpected exception in asyncore.loop()")

class NthControlClient(NthClient):

	def __init__(self, host, port, cmd, pretty=False):
		NthClient.__init__(self, host, port)
		self.done = False
		self.cmd = cmd
		self.pretty = pretty

	def onError (self, msg):
		print "NthControlClient error: ",msg
		raise asyncore.ExitNow
	   
	def onConnect (self):
		print "Connected..."
		try:
			self.push(str(self.cmd))
		except:
			print("Exception in push: "+format_exc())

	def onClose (self):
		print "NthControlClient: closed prematurely?"
		raise asyncore.ExitNow

	def onData (self, data):
		try:
			x = xmldom.parseString(data)
			if self.pretty:
				print(x.toprettyxml())
			else:
				print(x.toxml())
		except:
			print "ERROR in parsing xml: %s" % format_exc()
			print "Raw data was: %s" % data
		self.done = True


class NthMonitorClient(NthClient):

	def __init__(self, host, port):
		NthClient.__init__(self, host, port)

	def onError (self, msg):
		print "NthMonitorClient error: ",msg
		raise asyncore.ExitNow
	   
	def onConnect (self):
		print "Connected..."
		self.push("<next_event/>")

	def onClose (self):
		print "NthMonitorClient closed"
		raise asyncore.ExitNow

	def onData (self, data):
		try:
			x = xmldom.parseString(data)
		except:
			print "ERROR in parsing data=",data," : ",format_exc()
			return

		recognized = ("device","devices","error")

		if x.documentElement.tagName == "event":
			print data
			self.push("<next_event/>")
		elif x.documentElement.tagName in recognized:
			print data
		else:
			print "UNRECOGNIZED = ",x.toxml()


############################################################################

def doNthCommand(host,port,cmd):
	cc = NthControlClient(host,port,cmd)
	t = NthClientThread()
	t.start()
	while True:
		if cc.done:
			break
		if not t.isAlive():
			break
		try:
			t.join(2.0)
		except KeyboardInterrupt:
			print "GOT KEYBOARD INTERRUPT!"
			break

if __name__ == '__main__':

	print "INITIAL FRETUP"
	fev = FretUpEvent(22)
	print "INITIAL FRETUP=",fev


